unit MySQLClient;

{
  MySQL Native Delphi client (c)2003-2005 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr

  based on Mysql-direct library by Cristian Nicola

  v1.1  - bug fix for TMyField.Size &
  v1.2  - add PingTest for automatic reconnexion
  v2.0  - support for MySQL 4.1
  v2.01 - bug fix, empty password for 4.1
  v2.02 - bug fix, $ELSE/$ENDIF mistake in ExecSQL, (Michael)
  v2.03 - MySQL 4.1 can request an old password
}

{

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

// http://public.logicacmg.com/~redferni/mysql/MySQL-Protocol.html

{$DEFINE COMPRESS} // use ZLIB (deflated) compression (inbound data only)
{$DEFINE RECVBUF } // use an internal receive cache buffer
{$DEFINE SENDBUF } // use an internal sending cache buffer
{$DEFINE MYSQL41 } // support MySQL 4.1
{-DEFINE DUMP   } // dump to MySQLClient.log

{
  Tested on :
   MySQL 3.23.55-nt (Microsoft NT 4 SP 6)
   MySQL 4.0.15a (Novell Netware 6)
   MySQL 4.0.18-standard (RedHat 8)
   MySQL 4.1.9-standard (RedHat 7.2)


 Last change (use Search/Find "fix #x") :
  fix #1 : return code for MySQL 4.0.x
  fix #2 : size of non compressed data
  fix #3 : support empty password for 4.1
  fix #4 : $ELSE/$ENDIF mistake
  fix #5 : MySQL 4.1 request old encrypted password
}

interface

uses
 Windows,Classes,SysUtils,Winsock,Math,Dialogs{$IFDEF PRO},SqlTimSt{$ENDIF},zInflates;

type
 TMySQLFlag=(
  CLIENT_LONG_PASSWORD,     // new more secure passwords
  CLIENT_FOUND_ROWS,        // Found instead of affected rows
  CLIENT_LONG_FLAG,         // Get all column flags
  CLIENT_CONNECT_WITH_DB,   // One can specify db on connect
  CLIENT_NO_SCHEMA,         // Don't allow database.table.column
  CLIENT_COMPRESS,          // Can use compression protocol
  CLIENT_ODBC,              // Odbc client
  CLIENT_LOCAL_FILES,       // Can use LOAD DATA LOCAL
  CLIENT_IGNORE_SPACE,      // Ignore spaces before '('
 {$IFDEF MYSQL41}
  CLIENT_PROTOCOL_41,       // New 4.1 protocol
 {$ELSE}
  CLIENT_CHANGE_USER,       // ??
 {$ENDIF}
  CLIENT_INTERACTIVE,       // This is an interactive client
  CLIENT_SSL,               // Switch to SSL after handshake
  CLIENT_IGNORE_SIGPIPE,    // IGNORE sigpipes
  CLIENT_TRANSACTIONS,      // Client knows about transactions
  CLIENT_RESERVED,          // Old flag for 4.1 protocol
  CLIENT_SECURE_CONNECTION, // New 4.1 authentication
  CLIENT_MULTI_STATEMENTS,  // Enable/disable multi-stmt support
  CLIENT_MULTI_RESULTS      // Enable/disable multi-results
 );
 TMySQLFlags=set of TMySQLFlag;

 TMySQLCommand = (
  COM_SLEEP,          COM_QUIT,           COM_INIT_DB,
  COM_QUERY,          COM_FIELD_LIST,     COM_CREATE_DB,
  COM_DROP_DB,        COM_REFRESH,        COM_SHUTDOWN,
  COM_STATISTICS,     COM_PROCESS_INFO,   COM_CONNECT,
  COM_PROCESS_KILL,   COM_DEBUG,          COM_PING,
  COM_TIME,           COM_DELAYED_INSERT, COM_CHANGE_USER,
  COM_BINLOG_DUMP,    COM_TABLE_DUMP,     COM_CONNECT_OUT,
  COM_REGISTER_SLAVE, COM_PREPARE,        COM_EXECUTE,
  COM_LONG_DATA,      COM_CLOSE_STMT,     COM_RESET_STMT,
  COM_SET_OPTION
 );

 TMySQLFieldType = (
  dtDECIMAL,
  dtTINY,
  dtSHORT,
  dtLONG,
  dtFLOAT,
  dtDOUBLE,
  dtNULL,
  dtTIMESTAMP,
  dtLONGLONG,
  dtINT24,
  dtDATE,
  dtTIME,
  dtDATETIME,
  dtYEAR,
  dtNEWDATE,
  dtENUM=247,
  dtSET,
  dtTINY_BLOB,
  dtMEDIUM_BLOB,
  dtLONG_BLOB,
  dtBLOB,
  dtVAR_STRING,
  dtSTRING
 );

 // Field Flag Value Meaning
 TMySQLFieldFlag=(
  ffNOT_NULL,     //= 1, // Field can't be NULL
  ffPRI_KEY,      //= 2, // Field is part of a primary key
  ffUNIQUE_KEY,   //= 4, // Field is part of a unique key
  ffMULTIPLE_KEY, //= 8, // Field is part of a key
  ffBLOB,         //= 16,// Field is a blob
  ffUNSIGNED,     //= 32,// Field is unsigned
  ffZEROFILL,     //= 64,// Field is zerofill
  ffBINARY,       //= 128,// Field is binary
  ffENUM,         //= 256,// Field is an enum
  ffAUTO_INCREMENT, //=512,// Field is an autoincrement field
  ffTIMESTAMP,      //= 1024,/ Field is a timestamp
  ffSET,          //= 2048,// Field is a set
  ff_4096,
  ff_8192,
  ff_16384,
  ffNUM           // = 32768,// Field is num (for clients)
 );

 TMySQLFieldFlags=set of TMySQLFieldFlag;

 TMyField=class;

 TMySQLClient=class
 private
 // DoOpen
  fHostName   :string;
  fPort       :integer;
 // DoLogin
  fUserName   :string;
  fPassword   :string;
  fDataBase   :string;
  fHandle     :integer;
  fClientFlags:TMySQLFlags;
  fVersion    :byte;
  fServer     :string;
  fServerFlags:TMySQLFlags; // [CLIENT_LONG_FLAG..CLIENT_CONNECT_WITH_DB,CLIENT_COMPRESS]
  fScramble   :string;
  fFields     :TList;
  function SocketError:string;
  procedure SetCheck(Value:integer);
  procedure Error(msg:string);
  procedure CloseError(msg:string);
  procedure SetHandle(Value:integer);
  function inet(Host:pchar):integer;
  property Check:integer write SetCheck;
  procedure PingTest;
  procedure DoOpen;
  procedure DoLogin;
 protected
  fEof        :boolean;
  fBlockNumber:integer;
  fZCache :string;
  fAffectedRows:integer;
  fLastInsertID:integer;
 {$IFDEF RECVBUF}
  fRecvBuf:string;
  fRecvPos:integer;
  fRecvLen:integer;
 {$ENDIF}
 {$IFDEF SENDBUF}
  fSendBuf:string;
  fSendLen:integer;
  procedure SendBuffer;
 {$ENDIF}
  procedure Read(var Data; Size:integer);
  procedure Write(var Data; Size:integer);
  procedure BlockWrite(var Data; Size:integer; Extra:integer=0);
 {$IFDEF COMPRESS}
  procedure ZRead(var Data; Size:integer);
 {$ENDIF}
  function BlockRead:string;
  procedure Exec(Command:TMySQLCommand; Order:string);
  procedure ReadError;
  procedure SetFlag(Flag:TMySQLFlag; OnOff:boolean);
  function Crunch(Password:string):string;
  function Scramble(Password:string):string;
  function GetField(Index:integer):TMyField;
  property Handle:integer read fHandle write SetHandle;
  procedure Bind(const Data:string); virtual;
 public
  constructor Create; virtual;
  destructor Destroy; override;
  function Active:boolean;
  procedure Open(const AHostName:string; APort:integer=3306);
  procedure Close;
  procedure Ping;
  procedure Login(const AUserName,APassword:string; ADataBase:string='');
  procedure Select(DataBase:string);
  function PrimaryKey(ATable:string):string;
  function ExecSQL(const SQL:string):integer;
  procedure Fetch;
  procedure Flush; virtual;
  function FoundRows:integer;
  function FieldCount:integer;
  function FindField(const FieldName:string):TMyField;
  function FieldByName(const FieldName:string):TMyField;
  property Fields[Index:integer]:TMyField read GetField;
  property Eof:boolean read fEof;
  property AffectedRows:integer read fAffectedRows;
  property LastInsertID:integer read fLastInsertID;
 end;

 TMyField=class
 private
  fOwner :TMySQLClient;
  fName  :string;
  fTable :string;
 {$IFDEF MYSQL41}
  fCatalog :string;
  fDB      :string;
  fOrgTable:string;
  fOrgName :string;
 {$ENDIF}
  fSize  :integer;
  fKind  :TMySQLFieldType;
  fFlags :TMySQLFieldFlags;
  fPrec  :byte;
  fValue :string;
  function GetAsInteger:integer;
  function GetAsInt64:Int64;
  function GetAsDouble:double;
  function GetAsDateTime:TDateTime;
 {$IFDEF PRO}
  function GetAsTimeStamp:TSQLTimeStamp;
 {$ENDIF}
  function GetValue:string;
 public
  constructor Create(AOwner:TMySQLClient);
  destructor Destroy; override;
  property Name:string read fName write fName;
  property Table:string read fTable write fTable;
  property Size:integer read fSize write fSize;
  property Kind:TMySQLFieldType read fKind write fKind;
  property Flags:TMySQLFieldFlags read fFlags write fFlags;
  property Prec:byte read fPrec write fPrec;
  property AsString:string read fValue;
  property AsInteger:integer read GetAsInteger;
  property AsInt64:int64 read GetAsInt64;
  property AsDouble:double read GetAsDouble;
  property AsDateTime:TDateTime read GetAsDateTime;
 {$IFDEF PRO}
  property AsTimeStamp:TSQLTimeStamp read GetAsTimeStamp;
 {$ENDIF}
  property Value:string read GetValue;
 end;

function SQLFloatToStr(Float:Extended):string;
function SQLEncode(s:string):string;
function SQLDate(Date:TDateTime):string;
function SQLTime(Date:TDateTime):string;
function SQLDateTime(Date:TDateTime):string;
{$IFDEF PRO}
function SQLTimeStamp(const TimeStamp:TSQLTimeStamp):string;
{$ENDIF}
function Comma(ss:array of string):string;

implementation

resourcestring
 SSocketError='socket error';

{$IFDEF DUMP}
procedure Dump(method:string; data:pchar; size:integer);
const
 hx:array[0..$F] of char='0123456789ABCDEF';
var
 f:TextFile;
 s:string;
 c:string;
 i:integer;
begin
{$i-}
 AssignFile(f,'MySQLClient.log');
 Append(f);
 if IoResult<>0 then Rewrite(f);
 SetLength(s,4*size);
 SetLength(c,4*size); FillChar(c[1],4*Size,' ');
 for i:=0 to size-1 do begin
  s[4*i+1]:='#';
  s[4*i+2]:='$';
  s[4*i+3]:=hx[ord(data[i]) shr 4];
  s[4*i+4]:=hx[ord(data[i]) and $F];
  if (data[i]>#32)and(data[i]<#240) then c[4*i+3]:=data[i] else c[4*i+3]:='.';
 end;
 WriteLn(f,method,':'#13#10,s,#13#10,c,#13#10);
 CloseFile(f);
end;
{$ENDIF}

function SQLFloatToStr(Float:Extended):string;
var
 i:integer;
begin
 Result:=FloatToStr(Float);
 i:=pos(',',Result);
 if i>0 then Result[i]:='.';
end;

function SQLEncode(s:string):string;
var
 i,l,a:integer;
 c,d:char;
begin
 Result:=s;
 a:=0; l:=Length(s);
 for i:=1 to l do
  if s[i] in [#0,#10,#13,#26,'''','"','\'] then begin
   Result[i]:='\';
   inc(a);
  end;
 if a=0 then exit; // no escape needed
 i:=l+a;
 SetLength(Result,i); // reserve space for "\"
 FillChar(Result[l+1],a,' '); // initialiser les nouveaux caractres !
 while a>0 do begin
  c:=Result[l];
  if c='\' then begin // set by the above loop
   d:=s[l]; // original char
   case d of
    #00: d:='0';
    #10: d:='n';
    #13: d:='r';
    #26: d:='Z';
   end;
   Result[i]:=d; dec(i);
   dec(a);
  end;
  Result[i]:=c; dec(i);
  dec(l);
 end;
end;

procedure Int2StrMod(p:pchar; i,m:integer);
begin
 repeat
  inc(p^,i div m);
  i:=i mod m;
  m:=m div 10;
  inc(p);
 until (i=0)or(m=0);
end;

function SQLDate(Date:TDateTime):string;
var
 y,m,d:word;
begin
 DecodeDate(Date,y,m,d);
 Result:='0000-00-00';
 Int2StrMod(@Result[1],y,1000);
 Int2StrMod(@Result[6],m,10);
 Int2StrMod(@Result[9],d,10);
end;

function SQLTime(Date:TDateTime):string;
var
 h,m,s,sc:word;
begin
 DecodeTime(Date,h,m,s,sc);
 Result:='00:00:00';
 Int2StrMod(@Result[1],h,10);
 Int2StrMod(@Result[4],m,10);
 Int2StrMod(@Result[7],s,10);
end;

function SQLDateTime(Date:TDateTime):string;
begin
 Result:=SQLDate(Date)+' '+SQLTime(Date);
end;
{$IFDEF PRO}
function SQLTimeStamp(const TimeStamp:TSQLTimeStamp):string;
begin   //YYYYMMDDHHNNSS
 Result:='00000000000000';
 Int2StrMod(@Result[ 1],TimeStamp.Year,1000);
 Int2StrMod(@Result[ 5],TimeStamp.Month,10);
 Int2StrMod(@Result[ 7],TimeStamp.Day,10);
 Int2StrMod(@Result[ 9],TimeStamp.Hour,10);
 Int2StrMod(@Result[11],TimeStamp.Minute,10);
 Int2StrMod(@Result[13],TimeStamp.Second,10);
end;
{$ENDIF}
function Comma(ss:array of string):string;
var
 i:integer;
 s:string;
begin
 s:='';
 Result:='';
 for i:=Low(ss) to High(ss) do begin
  if ss[i]<>'' then begin
   Result:=Result+s+ss[i];
   s:=', ';
  end;
 end;
end;

// MySQL pre-4.1
procedure hash(s:string; var res0,res1:longint);
var
 nr1,add,nr2,tmp:int64;
 i:longint;
 e1:int64;
begin
 nr1:=1345345333;
 add:=7;
 nr2:=$12345671;
 for i:=1 to length(s) do begin
  if (s[i]=' ')or(s[i]=#9) then continue;
  tmp:=ord(s[i]);
  e1 :=(((nr1 and 63)+add)*tmp)+(nr1 shl 8);
  nr1:=nr1 xor e1;
 {$IFOPT Q+}{$DEFINE KEEPQ}{$ENDIF}{$Q-} // thanks Jiri Barton
  nr2:=nr2+((nr2 shl 8) xor nr1);
 {$IFDEF KEEPQ}{$Q+}{$UNDEF KEEPQ}{$ENDIF} // thanks Jiri Barton
  inc(add,tmp);
 end;
 res0:=nr1 and $7fffffff;
 res1:=nr2 and $7fffffff;
end;

// MySQL 4.1
type
 TSHA1Context=record
  Size :integer;
  Hash :array[0..4] of cardinal; // 20 bytes
  Index:integer;
  Block:array[0..63] of byte;
 end;

procedure SHA1Reset(var Context:TSHA1Context);
begin
 Context.Size:=0;
 Context.Hash[0]:=$67452301;
 Context.Hash[1]:=$EFCDAB89;
 Context.Hash[2]:=$98BADCFE;
 Context.Hash[3]:=$10325476;
 Context.Hash[4]:=$C3D2E1F0;
 Context.Index:=0;
end;

function SHA1CircularShift(bits,data:cardinal):cardinal;
begin
 Result:=(data shl bits) or (data shr (32-bits));
end;

procedure SHA1ProcessBlock(var Context:TSHA1Context);
const
 K:array[0..3] of cardinal=($5A827999,$6ED9EBA1,$8F1BBCDC,$CA62C1D6);
var
 W:array[0..79] of cardinal;
 t:integer;
 index:integer;
 A,B,C,D,E:cardinal;
 temp:cardinal;
begin
 // Initialize the first 16 words in the array W
 for t:=0 to 15 do begin
  index:=4*t;
  W[t]:=Context.Block[index  ] shl 24
     or Context.Block[index+1] shl 16
     or Context.Block[index+2] shl  8
     or Context.Block[index+3];
 end;
 for t:=16 to 79 do begin
  W[t]:=SHA1CircularShift(1,W[t-3] xor W[t-8] xor W[t-14] xor W[t-16]);
 end;
 A:=Context.Hash[0];
 B:=Context.Hash[1];
 C:=Context.Hash[2];
 D:=Context.Hash[3];
 E:=Context.Hash[4];
 for t:=0 to 19 do begin
  temp:=SHA1CircularShift(5,A)+((B and C) or ((not B) and D)) + E + W[t] + K[0];
  E:=D;
  D:=C;
  C:=SHA1CircularShift(30,B);
  B:=A;
  A:=temp;
 end;
 for t:=20 to 39 do begin
  temp:=SHA1CircularShift(5,A)+(B xor C xor D) + E + W[t] + K[1];
  E:=D;
  D:=C;
  C:=SHA1CircularShift(30,B);
  B:=A;
  A:=temp;
 end;
 for t:=40 to 59 do begin
  temp:=SHA1CircularShift(5,A)+((B and C) or (B and D) or (C and D)) + E + W[t] + K[2];
  E:=D;
  D:=C;
  C:=SHA1CircularShift(30,B);
  B:=A;
  A:=temp;
 end;
 for t:=60 to 79 do begin
  temp:=SHA1CircularShift(5,A)+(B xor C xor D) + E + W[t] + K[3];
  E:=D;
  D:=C;
  C:=SHA1CircularShift(30,B);
  B:=A;
  A:=temp;
 end;
 inc(Context.Hash[0],A);
 inc(Context.Hash[1],B);
 inc(Context.Hash[2],C);
 inc(Context.Hash[3],D);
 inc(Context.Hash[4],E);
 Context.Index:=0;
end;

procedure SHA1Input(var Context:TSHA1Context; Data:string);
var
 i:integer;
begin
 for i:=1 to Length(Data) do begin
  Context.Block[Context.Index]:=ord(Data[i]);
  inc(Context.Size);
  inc(Context.Index);
  if Context.Index=64 then SHA1ProcessBlock(Context);
 end;
end;

procedure SHA1PadMessage(var Context:TSHA1Context);
var
 i:integer;
begin
 i:=Context.Index;
 Context.Block[i]:=$80;
 inc(i);
 if i>56 then begin
  FillChar(Context.Block[i],64-i,0);
  Context.Index:=64;
  SHA1ProcessBlock(Context);
  FillChar(Context.Block[0],56,0);
 end else begin
  FillChar(Context.Block[i],56-i,0);
 end;
 Context.Index:=56;
 // Store the message length as the last 8 octets
 Context.Block[56]:=0;
 Context.Block[57]:=0;
 Context.Block[58]:=0;
 Context.Block[59]:=Context.Size shr 29;
 Context.Block[60]:=Context.Size shr 21;
 Context.Block[61]:=Context.Size shr 13;
 Context.Block[62]:=Context.Size shr  5;
 Context.Block[63]:=Context.Size shl  3;
 SHA1ProcessBlock(Context);
end;

function SHA1Result(var Context:TSHA1Context):string;
var
 i:integer;
begin
 SHA1PadMessage(Context);
 Context.Size:=0;
 SetLength(Result,20);
 for i:= 0 to 19 do
  Result[i+1]:=chr(Context.Hash[i shr 2] shr  (8 * ( 3 - ( i and 3 ) )));
end;

function VarSize(const Data:string; var Index:integer):integer;
var
 Code:integer;
begin
 if Length(Data)<Index then
  Result:=-1
 else begin
  Code:=ord(Data[Index]); inc(Index);
  case Code of
   251 : begin Result:=-1; exit; end;
   252 : begin Result:=pword(@Data[Index])^; inc(Index,2); end;
   253 : begin Result:=pinteger(@Data[Index])^ and $FFFFFF; inc(Index,3); end;
   254 : begin
          if (pinteger(@Data[Index])^)<>0 then raise Exception.Create('BIG DATA not supported');
          Result:=pinteger(@Data[Index+4])^;
          inc(Index,8);
         end;
   else  begin Result:=Code; exit; end;
  end;
 end;
end;

function VarData(const Data:string; var Index:integer):string;
var
 Len:integer;
begin
 Len:=VarSize(Data,Index);
 if Len<=0 then
  Result:=''
 else begin
  Result:=Copy(Data,Index,Len);
  inc(Index,Len);
 end;
end;

//----------------------------------------------------------------------------//

constructor TMySQLClient.Create;
begin
 inherited;
 fHandle:=INVALID_SOCKET;
 fFields:=TList.Create;
end;

destructor TMySQLClient.Destroy;
begin
 Close;
 fFields.Free;
 inherited;
end;

procedure TMySQLClient.Error(msg:string);
begin
 raise Exception.Create(msg);
end;

procedure TMySQLClient.CloseError(msg:string);
begin
 Close;
 Error(msg);
end;

function TMySQLClient.Crunch(Password:string):string;
var
 hp0,hp1:longint;
 hm0,hm1:longint;
 maxValue,seed, seed2 :int64;
 dRes: double;
 i:longint;
 e:byte;
 len1:longint;
 Scramble:string;
begin
 if (Password='') then
  Result:=''
 else begin
  Scramble:=copy(fScramble,1,8); // fix #5
  len1:=Length(Scramble);
  SetLength(Result,len1);
  hash(Password,hp0,hp1);
  hash(Scramble,hm0,hm1);
  maxValue:= $3FFFFFFF;
  seed  := ( hp0 xor hm0 )mod maxValue ;
  seed2 := ( hp1 xor hm1 )mod maxValue ;
  for i:=1 to len1 do begin
   seed  := ( seed * 3 + seed2  )mod maxValue ;
   seed2 := ( seed + seed2 + 33 )mod maxValue ;
   dRes := Seed / maxValue;
   result[i] := char( Floor( dRes * 31 ) + 64 );
  end;
  seed:=(seed * 3 + seed2) mod maxValue ;
  dRes:= Seed / maxValue;
  e := floor( dRes * 31 );
  for i := 1 to len1 do result[i]:=chr( byte(result[i]) xor e);
 end;
end;

function XorCrypt(const s1,s2:string):string;
var
 l,i:integer;
 a,b,c:^byte;
begin
 l:=Length(s1);
 SetLength(Result,l);
 a:=@s1[1];
 b:=@s2[1];
 c:=@Result[1];
 for i:=1 to l do begin
  c^:=a^ xor b^;
  inc(a);
  inc(b);
  inc(c);
 end;
end;

function TMySQLClient.Scramble(Password:string):string;
var
 SHA1Context:TSHA1Context;
 Stage1,Stage2:string;
begin
 // stage 1: hash password
 SHA1Reset(SHA1Context);
 SHA1Input(SHA1Context, Password);
 Stage1:=SHA1Result(SHA1Context);
 // stage 2: hash stage 1; note that hash_stage2 is stored in the database
 SHA1Reset(SHA1Context);
 SHA1Input(SHA1Context, Stage1);
 Stage2:=SHA1Result(SHA1Context);
 // create crypt string as sha1(message, hash_stage2)
 SHA1Reset(SHA1Context);
 SHA1Input(SHA1Context, fScramble);
 SHA1Input(SHA1Context, Stage2);
 Stage2:=SHA1Result(SHA1Context);
 
 Result:=XorCrypt(Stage2,Stage1);
end;

function TMySQLClient.SocketError:string;
var
 Value:integer;
begin
 Value:=WSAGetLastError;
 case Value of
  WSAECONNRESET : Result:='WSAECONNRESET : The virtual circuit was reset by the remote side';
  else Result:='WINSOCK Error #'+IntToStr(Value)+' (WSABASEERR+'+IntToStr(Value-WSABASEERR)+')';
 end;
end;

procedure TMySQLClient.SetCheck(Value:integer);
begin
 if Value=SOCKET_ERROR then Error(SocketError);
end;

procedure TMySQLClient.SetHandle(Value:integer);
begin
 if Value=INVALID_SOCKET then Error(SSocketError);
 fHandle:=Value;
end;

procedure TMySQLClient.SetFlag(Flag:TMySQLFlag; OnOff:boolean);
begin
 if OnOff then Include(fClientFlags,Flag) else Exclude(fClientFlags,Flag);
end;

function TMySQLClient.Active:boolean;
begin
 Result:=(fHandle<>INVALID_SOCKET);
end;

function TMySQLClient.inet(Host:pchar):integer;
var
 PHost:PHostEnt;
begin
 Result:=inet_addr(Host);
 if Result=INADDR_NONE then begin
  PHost:=gethostbyname(Host);
  if PHost=nil then Error('Host not found');
  Result:=integer(pointer(PHost.h_addr^)^);
 end;
end;

procedure TMySQLClient.Open(const AHostName:string; APort:integer=3306);
begin
 fHostName:=AHostName;
 fPort:=APort;
 DoOpen;
end;

procedure TMySQLClient.DoOpen;
var
 Sock:TSockAddr;
 Data:string;
 i,l:integer;
begin
 Close;
 Handle:=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
 try
 {$IFDEF RECVBUF}
  l:=SizeOf(i);
  if getsockopt(fHandle,SOL_SOCKET,SO_RCVBUF,pchar(@i),l)=SOCKET_ERROR then i:=1024;
  SetLength(fRecvBuf,i);
  fRecvLen:=0;
 {$ENDIF}
 {$IFDEF SENDBUF}
  l:=SizeOf(i);
  if getsockopt(fHandle,SOL_SOCKET,SO_SNDBUF,pchar(@i),l)=SOCKET_ERROR then
   i:=512
  else
   if i>512 then i:=512;
  SetLength(fSendBuf,i);
  fSendLen:=0;
 {$ENDIF}
  FillChar(Sock,SizeOf(Sock),0);
  Sock.sin_family:=AF_INET;
  Sock.sin_port:=htons(fPort);
  Sock.sin_addr.S_addr:=inet(pchar(fHostName));
  Check:=connect(fHandle,Sock,SizeOf(Sock));
  fServerFlags:=[];
  fBlockNumber:=0;
  Data:=BlockRead;
  fVersion :=ord(Data[1]);   // 1st byte = protocol version (10)
  if fVersion<>10 then Error('Invalide version');
  fServer:=pchar(@Data[2]);  // server string#0
  i:=1+Length(fServer)+1+5;  // (int)threadID
  fScramble:=Copy(Data,i,8); // pre-4.1 !
  inc(i,9);
  fServerFlags:=[];
  if Length(Data)>=i+2 then begin
   Move(Data[i],fServerFlags,2); inc(i,2); // 2 bytes
  {$IFDEF MYSQL41}
   // new prococol with 16 extra bytes
   if Length(Data)>=i+16 then begin
    // mysql->server_language=end[2]; <- Data[i]
    // mysql->server_status=uint2korr(end+3); <- Data[i+1..i+2]
    inc(i,16);
    if Length(Data)>=i+12 then begin
     fScramble:=fScramble+Copy(Data,i,12); // up to 20 bytes
    end;
   end;
  {$ENDIF}
  end;
  fEof:=True;
 except
  closesocket(fHandle);
  fHandle:=0;
  raise;
 end;
end;

procedure TMySQLClient.Close;
var
 i:integer;
begin
 if fHandle<>INVALID_SOCKET then begin
  for i:=fFields.Count-1 downto 0 do TMyField(fFields[i]).Free;
  closesocket(fHandle);
  fHandle:=INVALID_SOCKET;
  fEof:=True;
 end;
end;

procedure TMySQLClient.ReadError;
var
 Data:string;
begin
 Data:=BlockRead;
 if copy(Data,1,3)<>#0#0#0 then Error('Unknow error from server'); // fix #1
end;

procedure TMySQLClient.Ping;
begin
 Exec(COM_PING,'');
 ReadError;
end;

procedure TMySQLClient.Login(const AUserName,APassword:string; ADataBase:string='');
begin
 fUserName:=AUserName;
 fPassword:=APassword;
 fDataBase:=ADataBase;
 DoLogin;
end;

procedure TMySQLClient.DoLogin;
var
 Data:string;
 Pass:string;
 Flags:TMySQLFlags;
begin
 fClientFlags:=[CLIENT_LONG_PASSWORD,CLIENT_LONG_FLAG,CLIENT_LOCAL_FILES,CLIENT_TRANSACTIONS];
 if (fDataBase<>'')and(CLIENT_CONNECT_WITH_DB in fServerFlags) then Include(fClientFlags,CLIENT_CONNECT_WITH_DB);
 Flags:=fClientFlags;
{$IFDEF COMPRESS}
 if (CLIENT_COMPRESS in fServerFlags) then Include(Flags,CLIENT_COMPRESS);
{$ENDIF}
{$IFDEF MYSQL41}
 if (CLIENT_PROTOCOL_41 in fServerFlags) then begin
  Flags:=Flags+[CLIENT_PROTOCOL_41,CLIENT_SECURE_CONNECTION];
  SetLength(Data,32); FillChar(Data[1],32,0);
  Move(Flags,Data[1],4);
  Data[8]:=#$40; // $40000000 max allowed packet
  Data[9]:=#8;  // CHARSET
  if fPassword='' then // #fix 3
   Pass:=#0
  else
   Pass:=#20+Scramble(fPassword);
 end else
{$ENDIF}
 begin
 // send : Flags (2 bytes), MAX_ALLOWED_PACKET (3 bytes), UserZ, PassZ, DataBaseZ
  Data:=#0#0+#0#0#0;
  Pass:=Crunch(fPassword)+#0;
  Move(Flags,Data[1],2);
 end;
 Data:=Data+fUserName+#0+Pass+fDataBase+#0;
 BlockWrite(Data[1],Length(Data));
// ReadError;
 Data:=BlockRead;
 if Data=#254 then begin  // fix #5
  Data:=Crunch(fPassword)+#0;
  BlockWrite(Data[1],Length(Data));
  Data:=BlockRead;
 end;
 if copy(Data,1,3)<>#0#0#0 then Error('Unknow error from server');
 fClientFlags:=Flags; { Now we can set CLIENT_COMPRESS (see BlockRead) }
 if (fDataBase<>'')and(not (CLIENT_CONNECT_WITH_DB in fServerFlags)) then Select(fDataBase);
end;

procedure TMySQLClient.Select(DataBase:string);
begin
 Exec(COM_INIT_DB,DataBase);
 ReadError;
 fDataBase:=DataBase; // for DoLogin
end;

procedure TMySQLClient.Flush;
var
 i:integer;
 s:string;
begin
// Flush previous SQL !!!
 while not fEof do begin
  {Fetch}
  s:=BlockRead;
  fEof:=(s='')or(s[1]=#254);
 end;
 for i:=fFields.Count-1 downto 0 do TMyField(fFields[i]).Free;
end;


function TMySQLClient.FoundRows: integer;
var
 s:string;
 x:integer;
begin
 Flush;
 Exec(COM_QUERY,'SELECT FOUND_ROWS()');
 x:=1;
 s:=BlockRead;
 if VarSize(s,x)<>1 then Error('FOUND_ROWS() error !');
 s:=BlockRead; // Field Desc
 if BlockRead[1]<>#254 then Error('FoundRows error #1');
 x:=1;
 s:=BlockRead; // Value
 Result:=StrToInt(VarData(s,x));
 if BlockRead[1]<>#254 then Error('FoundRows error #2');
end;

function TMySQLClient.PrimaryKey(ATable:string):string;
begin
 Result:='';
 ExecSQL('DESC '+ATable);
 while not fEof do begin
  if Fields[3].AsString='PRI' then begin
   if Result<>'' then Result:=Result+', ';
   Result:=Result+Fields[0].AsString;
  end;
  Fetch;
 end;
 Flush;
end;

function TMySQLClient.ExecSQL(const SQL:string):integer;
var
 Data:string;
 i,FieldCount:integer;
 Field:TMyField;
 x:integer;
begin
 Flush;
{ en cas de coupure de communication entre deux requtes on peut esprer se reconnecter }
 PingTest;
 Exec(COM_QUERY,SQL);
 try
  Data:=BlockRead;
 except
  on e:Exception do raise Exception.Create(e.Message+#13#10+SQL);
 end;
 x:=1;
 FieldCount:=VarSize(Data,x);
 if FieldCount=0 then begin
 // INSERT, UPDATE...no data returned
  fAffectedRows:=VarSize(Data,x); // Affected Rows
  fLastInsertID:=VarSize(Data,x); // Insert ID
  Result:=fLastInsertID;
  // Status   (Word)
  // Warnings (Word) <-- v4.1
  // Message  (AsciiZ)
 end else begin
 // SELECT with returned data
  Result:=0;
  for i:=0 to FieldCount-1 do begin
   Data:=BlockRead;
   x:=1;
   Field:=TMyField.Create(Self);
  {$IFDEF MYSQL41}
   if (CLIENT_PROTOCOL_41 in fServerFlags) then begin
    Field.fCatalog :=VarData(Data,x); // 0
    Field.fDB      :=VarData(Data,x); // 1
    Field.fTable   :=VarData(Data,x); // 2
    Field.fOrgTable:=VarData(Data,x); // 3
    Field.fName    :=VarData(Data,x); // 4
    Field.fOrgName :=VarData(Data,x); // 5
    inc(x); // #12 (bytes follow)
    inc(x,2); // Charset
    Field.fSize    :=PInteger(@Data[x])^; inc(x,4);
    Field.fKind    :=TMySQLFieldType(Data[x]); inc(x); //
    Field.fFlags   :=TMySQLFieldFlags(PWord(@Data[x+1])^); inc(x,2);
    Field.fPrec    :=ord(Data[x]);
    // Data[x+1] = Default value ASCIIZ
   end else
  {$ENDIF} // fix #4
   begin
    Field.fTable:=VarData(Data,x);
    Field.fName :=VarData(Data,x);
    inc(x); // #3
    Field.fSize :=PInteger(@Data[x])^ and $FFFFFF;
    inc(x,4); // Size #3
    Field.fKind :=TMySQLFieldType(Data[x]);
    inc(x);
    if Data[x]=#2 then begin
     word(Field.fFlags):=ord(Data[x+1]);
     inc(x,2);
    end else begin
     word(Field.fFlags):=PWord(@Data[x+1])^ ;
     inc(x,3);
    end;
    if x<>Length(Data) then Error('Error while reading field def');
    Field.fPrec:=ord(Data[x]);
   end;
   // fix #4
  end;
  Data:=BlockRead;
  if (Data='')or(Data[1]<>#254) then error('ExecSQL error');
  fEof:=False;
  Fetch;
 end;
end;

procedure TMySQLClient.Read(var Data; Size:integer);
var
 p:pchar;
 i:integer;
begin
 p:=@Data;
 while Size>0 do begin
{$IFDEF RECVBUF}
  if fRecvLen=0 then begin
   fRecvLen:=recv(fHandle,fRecvBuf[1],Length(fRecvBuf),0);
   if fRecvLen<=0 then CloseError('Read error '+SocketError);
  {$IFDEF DUMP}
   Dump('recv',@fRecvBuf[1],fRecvLen);
  {$ENDIF}
   fRecvPos:=0;
  end;
  i:=fRecvLen-fRecvPos;
  if i>Size then i:=Size;
  Move(fRecvBuf[fRecvPos+1],p[0],i);
  inc(fRecvPos,i);
  if fRecvPos=fRecvLen then fRecvLen:=0;
{$ELSE}
  i:=recv(fHandle,p[0],Size,0);
  if i<=0 then CloseError('Read error '+SocketError);
{$ENDIF}
  inc(p,i);
  dec(Size,i);
 end;
end;

procedure TMySQLClient.Write(var Data; Size:integer);
var
 p:pchar;
 i:integer;
begin
 p:=@Data;
 while Size>0 do begin
 {$IFDEF SENDBUF}
  i:=Length(fSendBuf)-fSendLen;
  if i>Size then i:=Size;
  move(p[0],fSendBuf[fSendLen+1],i);
  inc(fSendLen,i);
  if fSendLen=Length(fSendBuf) then SendBuffer;
 {$ELSE}
  i:=send(fHandle,p[0],Size,0);
  if i<=0 then CloseError('Write error '+SocketError);
 {$IFDEF DUMP}
  Dump('send',p[0],i);
 {$ENDIF}
 {$ENDIF}
  inc(p,i);
  dec(Size,i);
 end;
end;

{$IFDEF SENDBUF}
procedure TMySQLClient.SendBuffer;
var
 p:pchar;
 i:integer;
begin
 p:=@fSendBuf[1];
 while fSendLen>0 do begin
  i:=send(fHandle,p[0],fSendLen,0);
  if i<=0 then CloseError('Write error '+SocketError);
  {$IFDEF DUMP}
   Dump('send',@p[0],i);
  {$ENDIF}
  dec(fSendLen,i);
  inc(p,i);
 end;
end;
{$ENDIF}

procedure TMySQLClient.BlockWrite(var Data; Size:integer; Extra:integer=0);
var
 Head:integer;
 Comp:record
  RealSize:integer;
  CompSize:integer;
 end;
begin
{$IFDEF COMPRESS}
 if (CLIENT_COMPRESS in fClientFlags) then begin
  Comp.RealSize:=(fBlockNumber shl 24) or (SizeOf(Head)+Size+Extra);
  fBlockNumber:=Succ(fBlockNumber) and $FF;
  Comp.CompSize:=0; // do not compress...
  Write(Comp,7);    // CompSize = 3 bytes only
  Head:=Size+Extra; // BlockNumber = 0
 end else
{$ENDIF} 
 begin
  Head:=(fBlockNumber shl 24) or (Size+Extra);
  fBlockNumber:=Succ(fBlockNumber) and $FF;
 end;
 Write(Head,SizeOf(Head));
 Write(Data,Size);
end;
{$IFDEF COMPRESS}
procedure TMySQLClient.ZRead(var Data; Size:integer);
var
 Len :integer;
 Head:record
  CompSize:integer; // 3 bytes + BlockNumber
  RealSize:integer; // 3 bytes only !!
 end;
 Comp:string;
begin
 Len:=Length(fZCache);
 while Len<Size do begin
  Head.RealSize:=0; // 4th byte
  Read(Head,7);     // compression Header
  if Head.CompSize shr 24<>fBlockNumber then CloseError('ZProtocol error #1');
  fBlockNumber:=Succ(fBlockNumber) and $FF;
  Head.CompSize:=Head.CompSize and $FFFFFF; // fix #2
  if Head.RealSize=0 then begin // no compression
   SetLength(fZCache,Len+Head.CompSize);
   Read(fZCache[Len+1],Head.CompSize);
   inc(Len,Head.CompSize);
  end else begin
   SetLength(Comp,Head.CompSize);
   Read(Comp[1],Head.CompSize);
   SetLength(fZCache,Len+Head.RealSize);
  // Check DEFLATED compression...hope it's ok :}
   if (Comp[1]<>#$78)or(Comp[2]<>#$9C) then CloseError('ZProtocol error #2');
   Inflate(Comp[3],fZCache[Len+1]);
   inc(Len,Head.RealSize);
  end;
 end;
 Move(fZCache[1],Data,Size);
 if Size=Len then fZCache:='' else Delete(fZCache,1,Size);
end;
{$ENDIF}
function TMySQLClient.BlockRead:string;
var
 Size:integer; // 3 bytes + BlockNumber
begin
{$IFDEF SENDBUF}
 SendBuffer; // make sure that the request is send before ;D
{$ENDIF}
{$IFDEF COMPRESS}
 if CLIENT_COMPRESS in fClientFlags then begin
  ZRead(Size,SizeOf(Size));
  Size:=Size and $FFFFFF; // 4.018a fix #3
  SetLength(Result,Size);
  ZRead(Result[1],Size);
 end else
{$ENDIF}
 begin
  Read(Size,SizeOf(Size));
  if Size shr 24<>fBlockNumber then CloseError('Protocol error #1');
  fBlockNumber:=Succ(fBlockNumber) and $FF;
  Size:=Size and $FFFFFF;
  SetLength(Result,Size);
  Read(Result[1],Size);
 end;
 if Result='' then CloseError('MySQL Error : empty response');
 if Result[1]=#255 then CloseError('MySQL Error #'+IntToStr(pword(@Result[2])^)+' : '+Copy(Result,4,Size));
end;

procedure TMySQLClient.Exec(Command:TMySQLCommand; Order:string);
var
 Extra:integer;
begin
 fBlockNumber:=0;
 Extra:=Length(Order);
 BlockWrite(Command,1,Extra);
 Write(Order[1],Extra);
end;

function TMySQLClient.FieldCount:integer;
begin
 Result:=fFields.Count;
end;

function TMySQLClient.FindField(const FieldName:string):TMyField;
var
 i:integer;
begin
 i:=fFields.Count-1;
 while (i>=0) do begin
  Result:=fFields[i];
  if AnsiCompareText(Result.Name,FieldName)=0 then exit;
  dec(i);
 end;
 Result:=nil;
end;

function TMySQLClient.FieldByName(const FieldName:string):TMyField;
begin
 Result:=FindField(FieldName);
 if Result=nil then raise Exception.Create('Field "'+FieldName+'" not found');
end;

function TMySQLClient.GetField(index:integer):TMyField;
begin
 Result:=fFields[Index];
end;

procedure TMySQLClient.Fetch;
begin
 Bind(BlockRead);
end;

procedure TMySQLClient.Bind(const Data:string);
var
  i,x:integer;
begin
 if (Data='')or(Data[1]=#254) then begin
  fEof:=True;
 end else begin
  x:=1;
  for i:=0 to fFields.Count-1 do TMyField(fFields[i]).fValue:=VarData(Data,x);
 end;
end;

procedure TMySQLClient.PingTest;
begin
// reprise impossible si le buffer n'est pas vide
{$IFDEF SENDBUF}if fSendLen<>0 then exit;{$ENDIF}
 Exec(COM_PING,'');
 try
  ReadError;
 except
  DoOpen;
  DoLogin;
  Beep;
 end;
end;

//----------------------------------------------------------------------------//
constructor TMyField.Create(AOwner:TMySQLClient);
begin
 fOwner:=AOwner;
 if fOwner<>nil then fOwner.fFields.Add(Self);
end;

destructor TMyField.Destroy;
begin
 if fOwner<>nil then fOwner.fFields.Remove(Self);
 inherited;
end;

function TMyField.GetAsDouble: double;
var
 s:string;
 i:integer;
begin
 if fValue='' then Result:=0 else begin
  s:=fValue;
  i:=pos('.',s);
  if i>0 then s[i]:=',';
  Result:=StrToFloat(s);
 end;
end;

function TMyField.GetAsInteger:integer;
begin
 if fValue='' then Result:=0 else Result:=StrToInt(fValue);
end;

function TMyField.GetAsInt64:int64;
begin
 if fValue='' then Result:=0 else Result:=StrToInt(fValue);
end;

function TMyField.GetAsDateTime:TDateTime;
//           1
//  1234567890123456789
// "YYYY-MM-DD HH:MM:SS"
var
 y,m,d,h,n,s:integer;
begin
 if Length(fValue)>=4  then y:=StrToInt(copy(fValue,1 ,4)) else y:=0;
 if Length(fValue)>=7  then m:=StrToInt(copy(fValue,6 ,2)) else m:=0;
 if Length(fValue)>=10 then d:=StrToInt(copy(fValue,9 ,2)) else d:=0;
 if Length(fValue)>=13 then h:=StrToInt(copy(fValue,12,2)) else h:=0;
 if Length(fValue)>=16 then n:=StrToInt(copy(fValue,15,2)) else n:=0;
 if Length(fValue)>=19 then s:=StrToInt(copy(fValue,18,2)) else s:=0;
 Result:=EncodeDate(y,m,d)+EncodeTime(h,n,s,0);
end;
{$IFDEF PRO}
function TMyField.GetAsTimeStamp: TSQLTimeStamp;
// YYYYMMDDHHMMSS
begin
 if fValue='' then
  FillChar(Result,SizeOf(Result),0)
 else begin
  Result.Year  :=StrToInt(copy(fValue, 1,4));
  Result.Month :=StrToInt(copy(fValue, 5,2));
  Result.Day   :=StrToInt(copy(fValue, 7,2));
  Result.Hour  :=StrToInt(copy(fValue, 9,2));
  Result.Minute:=StrToInt(copy(fValue,11,2));
  Result.Second:=StrToInt(copy(fValue,13,2));
  Result.Fractions:=0;
 end;
end;
{$ENDIF}
function TMyField.GetValue:string;
begin
 if fKind in [dtDATE..dtSTRING] then
  Result:=''''+SQLEncode(fValue)+''''
 else
  Result:=fValue;
end;

//----------------------------------------------------------------------------//
procedure WSAStartup;
var
 wsa:TWSAData;
begin
 Winsock.WSAStartup($101,wsa);
end;

initialization
 WSAStartup;
finalization
 WSACleanup;
end.
